import java.util.Collections;
import java.util.LinkedList;

// BEGIN SOLUTION

void delaunayfy(DCEL triangulation) {
  boolean allLegal = false;
  do {
    allLegal = true;
    for (DCEL.HalfEdge e : triangulation.halfEdges) {
      if (!e.isLegal()) {
        e.flip();
        allLegal = false;
        break;
      }
    }
  } while (!allLegal);
}


// END SOLUTION

/**
 * Compute the convex hull of S using the incremental algorithm. 
 */ 
DCEL triangulationIncremental(ArrayList<Point2d> S) {
  
  // Sort S by x-coordinate using the Point2dXComparator: 
  Collections.sort(S, new Point2dXComparator());
  
  // Initialize the convex hull of the first three points
  DCEL triangulation = getCCWTriangle(S.get(0), S.get(1), S.get(2));
  
  // Now loop over the remaining points in S starting from k=3
  // and add each one to the convex hull to compute the hull
  // via the incremental algorithm. 
  
  for (int k = 3; k < S.size(); k++) {
    Point2d pk = S.get(k);
    DCEL.HalfEdge he, upperTangent = null, lowerTangent = null; 
    
    ArrayList<DCEL.HalfEdge> tangents = new ArrayList<DCEL.HalfEdge>();
    he = triangulation.outerFace.incidentEdge;
    do {
      if (isSameSide(pk, he.origin.coord, he.next.origin.coord, he.prev.origin.coord)) {
        tangents.add(he);
      }
      he = he.next;
    } while (he != triangulation.outerFace.incidentEdge);
    
    if (tangents.size() == 2) {
      DCEL.Vertex newVertex = triangulation.new Vertex(pk);
      
      if (isLeftHandTurn(pk, tangents.get(0).origin.coord, tangents.get(0).next.origin.coord)) {
        upperTangent = tangents.get(0);
        lowerTangent = tangents.get(1);
      } else {
        upperTangent = tangents.get(1);
        lowerTangent = tangents.get(0);
      }
      
      // Now hook up the DCEL edges. 
      DCEL.Face newFace = triangulation.newFace();
      
      DCEL.HalfEdge upper1 = triangulation.newHalfEdge(newVertex, newFace);
      DCEL.HalfEdge upper2 = triangulation.newHalfEdge(upperTangent.origin, triangulation.outerFace);
      
      DCEL.HalfEdge lower1 = triangulation.newHalfEdge(lowerTangent.origin, newFace);
      DCEL.HalfEdge lower2 = triangulation.newHalfEdge(newVertex, triangulation.outerFace);
      
      upper1.makeTwins(upper2);
      lower1.makeTwins(lower2);
      
      upperTangent.prev.makeNext(upper2);
      upper2.makeNext(lower2);
      
      upper1.makeNext(upperTangent);
      lowerTangent.prev.makeNext(lower1);
      lower1.makeNext(upper1);
      
      lower2.makeNext(lowerTangent);
      
      for (DCEL.HalfEdge edge : triangulation.getCycleFrom(upper1)) {
        edge.face = newFace;
      }
      newFace.incidentEdge = upper1;
      triangulation.outerFace.incidentEdge = lower2;
      
      while (upper1.next.next.next != upper1) {
        triangulation.splitFace(upper1, upper1.next.next);
        upper1 = upper1.prev.twin;
      }
      
    } else {
    
    }
  }
  
  return triangulation;
}

DCEL getCCWTriangle(Point2d a, Point2d b, Point2d c) {
 
  DCEL tri = new DCEL(true, false);
  
  DCEL.Vertex v1;
  DCEL.Vertex v2;
  DCEL.Vertex v3;
  
  if (isLeftHandTurn(a, b, c)) {
    v1 = tri.newVertex(a);
    v2 = tri.newVertex(b);
    v3 = tri.newVertex(c);
  } else {
    v1 = tri.newVertex(a);
    v2 = tri.newVertex(c);
    v3 = tri.newVertex(b);
  }
  
  // Create the inner face
  DCEL.Face inner = tri.newFace();
  
  // Create the inner half-edges
  DCEL.HalfEdge he12 = tri.newHalfEdge(v1, inner);
  DCEL.HalfEdge he23 = tri.newHalfEdge(v2, inner);
  DCEL.HalfEdge he31 = tri.newHalfEdge(v3, inner);
  
  // Connect the next/prev pointers on the inner face
  he12.makeNext(he23);
  he23.makeNext(he31);
  he31.makeNext(he12);
  
  // Create the outer half-edges
  DCEL.HalfEdge he21 = tri.newHalfEdge(v2, tri.outerFace);
  DCEL.HalfEdge he13 = tri.newHalfEdge(v1, tri.outerFace);
  DCEL.HalfEdge he32 = tri.newHalfEdge(v3, tri.outerFace);
  
  // Connect the next/perv pointers on the outer face
  he21.makeNext(he13);
  he13.makeNext(he32);
  he32.makeNext(he21);
  
  // Set the twins
  he12.makeTwins(he21);
  he23.makeTwins(he32);
  he13.makeTwins(he31);
  
  return tri;
}