import java.util.Random;

DCEL triangulation;
DCEL dual;

DCEL.Vertex selectedVertex = null; 

boolean showDelaunayEdges = true;
boolean showVoronoiEdges = true;

int nPoints = 20;

ArrayList<Point2d> S;

int to_x(double coordX) {
  return (int)Math.round(coordX + (width / 2));
}

double from_x(double screenX) {
  return screenX - (width / 2);
}

int to_y(double coordY) {
  return (int)Math.round((height/2) - coordY);
}

double from_y(int screenY) {
  return (height/2) - screenY;
}

void setup() {
  size(800, 600);
  smooth(4);
  background(255);

  triangulateRandomPointSet(473, nPoints);
  delaunayfy(triangulation);
  dual = triangulation.dual();
}

double perturb(Random r) {
  return (r.nextDouble() - 0.5)*0.0001;
}

void triangulateRandomPointSet(long seed, int n) {
  S = new ArrayList<Point2d>();
  triangulation = new DCEL(true, false);
  
  Random r = new Random(seed);
  for (int i = 0; i < n; i++) {
    S.add(new Point2d(r.nextDouble()*width - (width/2.0), r.nextDouble()*height - (height/2.0)));
  }
  
  triangulation = triangulationIncremental(S);
}


void retriangulate() {
  S = new ArrayList<Point2d>();
  
  for (DCEL.Vertex v : triangulation.vertices) {
    S.add(v.coord);
  }
  
  triangulation = triangulationIncremental(S);
}

void drawTriangulation() {
  
  // Draw edges
  if (showDelaunayEdges) {
    stroke(0, 0, 255);
    for (DCEL.HalfEdge e : triangulation.halfEdges) {
      line(to_x(e.origin.coord.x), to_y(e.origin.coord.y), 
              to_x(e.twin.origin.coord.x), to_y(e.twin.origin.coord.y));
    }
  }
  
  // Draw vertices
  fill(255);
  stroke(0);
  for (DCEL.Vertex v : triangulation.vertices) {
    ellipse(to_x(v.coord.x), to_y(v.coord.y), 8, 8);
  }
}

void drawDual() {
  
  if (dual == null) return;
  
  // Draw edges
  if (showVoronoiEdges) {
    stroke(255, 0, 0);
    for (DCEL.HalfEdge e : dual.halfEdges) {
      if (e.origin != dual.infinityVertex && e.twin.origin != dual.infinityVertex) 
        line(to_x(e.origin.coord.x), to_y(e.origin.coord.y), 
             to_x(e.twin.origin.coord.x), to_y(e.twin.origin.coord.y));
    }
  }
  
  // Draw vertices
  fill(255, 0, 0);
  stroke(255, 0, 0);
  for (DCEL.Vertex v : dual.vertices) {
    if (v != dual.infinityVertex) 
      ellipse(to_x(v.coord.x), to_y(v.coord.y), 8, 8);
  }
}

void draw() {
  background(255);
  Random rand = new Random(32);
  
  drawTriangulation();
  drawDual();
}

void keyReleased() {
  if (key == 'r') {
    triangulateRandomPointSet(System.currentTimeMillis(), nPoints);
    delaunayfy(triangulation);
    dual = triangulation.dual();
  } else if (key == '+' || key == '=') {
    nPoints ++;
    System.out.println(nPoints);
    triangulateRandomPointSet(System.currentTimeMillis(), nPoints);
    delaunayfy(triangulation);
    dual = triangulation.dual();
  } else if (key == '-' && nPoints > 3) {
    nPoints --;
    triangulateRandomPointSet(System.currentTimeMillis(), nPoints);
    delaunayfy(triangulation);
    dual = triangulation.dual();
  } else if (key == 'd') {
    showDelaunayEdges = !showDelaunayEdges; 
  } else if (key == 'v') {
    showVoronoiEdges = !showVoronoiEdges; 
  }
}

void keyPressed() {
}

void mousePressed() {
  for (DCEL.Vertex v : triangulation.vertices) {
    if (v.coord.distSqTo(new Point2d(from_x(mouseX), from_y(mouseY))) < 64) {
      selectedVertex = v;
      break;
    }
  }
}

void mouseDragged() {
  if (selectedVertex != null) {
    selectedVertex.coord.x = from_x(mouseX)+perturb(new Random());
    selectedVertex.coord.y = from_y(mouseY)+perturb(new Random());
    retriangulate();
    delaunayfy(triangulation);
    dual = triangulation.dual();
  }
}

void mouseReleased() {
  selectedVertex = null;
}