
/**
 * @return true if A-B-C constitutes a left-hand turn; false otherwise.
 */
boolean isLeftHandTurn(Point2d A, Point2d B, Point2d C) {
  return areaOfParallelogram(A, B, C) > 0;
}

/**
 * @return true if A-B-C constitutes a left-hand turn or collinear; false otherwise.
 */
boolean isLeftHandTurnOrOn(Point2d A, Point2d B, Point2d C) {
  return areaOfParallelogram(A, B, C) >= 0;
}

/**
 * @return true if the set of points A-B-C is collinear; false otherwise.
 */
boolean isCollinear(Point2d A, Point2d B, Point2d C) {
  return areaOfParallelogram(A, B, C) == 0;
}

/**
 * @return true if A-B-C constitutes a right-hand turn or collinear; false otherwise.
 */
boolean isRightHandTurn(Point2d A, Point2d B, Point2d C) {
  return areaOfParallelogram(A, B, C) < 0;
}

/**
 * @return true if A-B-C constitutes a right-hand turn; false otherwise.
 */
boolean isRightHandTurnOrOn(Point2d A, Point2d B, Point2d C) {
  return areaOfParallelogram(A, B, C) <= 0;
}

boolean isSameSide(Point2d A, Point2d B, Point2d C, Point2d D) {
  return (isLeftHandTurn(A, B, C) && isLeftHandTurn(A, B, D)) || 
         (isRightHandTurn(A, B, C) && isRightHandTurn(A, B, D));
}

/**
 * @return the signed area of the parallelogram with consecutive vertices B, A, C
 */
double areaOfParallelogram(Point2d A, Point2d B, Point2d C) {
    return (B.x - A.x) * (C.y - A.y) - (C.x - A.x) * (B.y - A.y);
}

/**
 * @return the signed area of the triangle with corners A, B, C.
 */
double areaOfTriangle(Point2d A, Point2d B, Point2d C) {
    return 0.5 * areaOfParallelogram(A, B, C);
}

/** 
 * Tests if point D is in the circle through A, B, C
 * Adapted from: https://www.cs.cmu.edu/afs/cs/project/quake/public/code/predicates.c
 */ 
boolean inCircle(Point2d A, Point2d B, Point2d C, Point2d D) {
  double adx, ady, bdx, bdy, cdx, cdy;
  double abdet, bcdet, cadet;
  double alift, blift, clift;

  adx = A.x - D.x;
  ady = A.y - D.y;
  bdx = B.x - D.x;
  bdy = B.y - D.y;
  cdx = C.x - D.x;
  cdy = C.y - D.y;

  abdet = adx * bdy - bdx * ady;
  bcdet = bdx * cdy - cdx * bdy;
  cadet = cdx * ady - adx * cdy;
  alift = adx * adx + ady * ady;
  blift = bdx * bdx + bdy * bdy;
  clift = cdx * cdx + cdy * cdy;

  return (alift * bcdet + blift * cadet + clift * abdet) > 0;
}